@@ -1,4 +1,4 @@
-// Copyright 2017-2019 The NATS Authors
+// Copyright 2017-2021 The NATS Authors
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
 // You may obtain a copy of the License at
@@ -24,7 +24,7 @@ import (
 	"sync/atomic"
 	"time"
 
-	"github.com/nats-io/jwt"
+	"github.com/nats-io/jwt/v2"
 )
 
 // FlagSnapshot captures the server options as specified by CLI flags at
@@ -50,9 +50,16 @@ type option interface {
 	// IsAuthChange indicates if this option requires reloading authorization.
 	IsAuthChange() bool
 
+	// IsTLSChange indicates if this option requires reloading TLS.
+	IsTLSChange() bool
+
 	// IsClusterPermsChange indicates if this option requires reloading
 	// cluster permissions.
 	IsClusterPermsChange() bool
+
+	// IsJetStreamChange inidicates a change in the servers config for JetStream.
+	// Account changes will be handled separately in reloadAuthorization.
+	IsJetStreamChange() bool
 }
 
 // noopOption is a base struct that provides default no-op behaviors.
@@ -70,10 +77,18 @@ func (n noopOption) IsAuthChange() bool {
 	return false
 }
 
+func (n noopOption) IsTLSChange() bool {
+	return false
+}
+
 func (n noopOption) IsClusterPermsChange() bool {
 	return false
 }
 
+func (n noopOption) IsJetStreamChange() bool {
+	return false
+}
+
 // loggingOption is a base struct that provides default option behaviors for
 // logging-related options.
 type loggingOption struct {
@@ -122,9 +137,11 @@ type debugOption struct {
 	newValue bool
 }
 
-// Apply is a no-op because logging will be reloaded after options are applied.
+// Apply is mostly a no-op because logging will be reloaded after options are applied.
+// However we will kick the raft nodes if they exist to reload.
 func (d *debugOption) Apply(server *Server) {
 	server.Noticef("Reloaded: debug = %v", d.newValue)
+	server.reloadDebugRaftNodes()
 }
 
 // logtimeOption implements the option interface for the `logtime` setting.
@@ -192,6 +209,10 @@ func (t *tlsOption) Apply(server *Server) {
 	server.Noticef("Reloaded: tls = %s", message)
 }
 
+func (t *tlsOption) IsTLSChange() bool {
+	return true
+}
+
 // tlsTimeoutOption implements the option interface for the tls `timeout`
 // setting.
 type tlsTimeoutOption struct {
@@ -205,6 +226,17 @@ func (t *tlsTimeoutOption) Apply(server *Server) {
 	server.Noticef("Reloaded: tls timeout = %v", t.newValue)
 }
 
+// tlsPinnedCertOption implements the option interface for the tls `pinned_certs` setting.
+type tlsPinnedCertOption struct {
+	noopOption
+	newValue PinnedCertSet
+}
+
+// Apply is a no-op because the pinned certs will be reloaded after options are  applied.
+func (t *tlsPinnedCertOption) Apply(server *Server) {
+	server.Noticef("Reloaded: %d pinned_certs", len(t.newValue))
+}
+
 // authOption is a base struct that provides default option behaviors.
 type authOption struct {
 	noopOption
@@ -289,23 +321,28 @@ type clusterOption struct {
 }
 
 // Apply the cluster change.
-func (c *clusterOption) Apply(server *Server) {
+func (c *clusterOption) Apply(s *Server) {
 	// TODO: support enabling/disabling clustering.
-	server.mu.Lock()
+	s.mu.Lock()
 	tlsRequired := c.newValue.TLSConfig != nil
-	server.routeInfo.TLSRequired = tlsRequired
-	server.routeInfo.TLSVerify = tlsRequired
-	server.routeInfo.AuthRequired = c.newValue.Username != ""
+	s.routeInfo.TLSRequired = tlsRequired
+	s.routeInfo.TLSVerify = tlsRequired
+	s.routeInfo.AuthRequired = c.newValue.Username != ""
 	if c.newValue.NoAdvertise {
-		server.routeInfo.ClientConnectURLs = nil
+		s.routeInfo.ClientConnectURLs = nil
+		s.routeInfo.WSConnectURLs = nil
 	} else {
-		server.routeInfo.ClientConnectURLs = server.clientConnectURLs
+		s.routeInfo.ClientConnectURLs = s.clientConnectURLs
+		s.routeInfo.WSConnectURLs = s.websocket.connectURLs
 	}
-	server.setRouteInfoHostPortAndIP()
-	server.mu.Unlock()
-	server.Noticef("Reloaded: cluster")
+	s.setRouteInfoHostPortAndIP()
+	s.mu.Unlock()
+	if c.newValue.Name != "" && c.newValue.Name != s.ClusterName() {
+		s.setClusterName(c.newValue.Name)
+	}
+	s.Noticef("Reloaded: cluster")
 	if tlsRequired && c.newValue.TLSConfig.InsecureSkipVerify {
-		server.Warnf(clusterTLSInsecureWarning)
+		s.Warnf(clusterTLSInsecureWarning)
 	}
 }
 
@@ -515,7 +552,7 @@ type clientAdvertiseOption struct {
 // Apply the setting by updating the server info and regenerate the infoJSON byte array.
 func (c *clientAdvertiseOption) Apply(server *Server) {
 	server.mu.Lock()
-	server.setInfoHostPortAndGenerateJSON()
+	server.setInfoHostPort()
 	server.mu.Unlock()
 	server.Noticef("Reload: client_advertise = %s", c.newValue)
 }
@@ -531,6 +568,29 @@ func (a *accountsOption) Apply(s *Server) {
 	s.Noticef("Reloaded: accounts")
 }
 
+// For changes to a server's config.
+type jetStreamOption struct {
+	noopOption
+	newValue bool
+}
+
+func (a *jetStreamOption) Apply(s *Server) {
+	s.Noticef("Reloaded: JetStream")
+}
+
+func (jso jetStreamOption) IsJetStreamChange() bool {
+	return true
+}
+
+type ocspOption struct {
+	noopOption
+	newValue *OCSPConfig
+}
+
+func (a *ocspOption) Apply(s *Server) {
+	s.Noticef("Reloaded: OCSP")
+}
+
 // connectErrorReports implements the option interface for the `connect_error_reports`
 // setting.
 type connectErrorReports struct {
@@ -569,22 +629,109 @@ func (m *maxTracedMsgLenOption) Apply(server *Server) {
 	server.Noticef("Reloaded: max_traced_msg_len = %d", m.newValue)
 }
 
-// Reload reads the current configuration file and applies any supported
-// changes. This returns an error if the server was not started with a config
-// file or an option which doesn't support hot-swapping was changed.
+type mqttAckWaitReload struct {
+	noopOption
+	newValue time.Duration
+}
+
+func (o *mqttAckWaitReload) Apply(s *Server) {
+	s.Noticef("Reloaded: MQTT ack_wait = %v", o.newValue)
+}
+
+type mqttMaxAckPendingReload struct {
+	noopOption
+	newValue uint16
+}
+
+func (o *mqttMaxAckPendingReload) Apply(s *Server) {
+	s.mqttUpdateMaxAckPending(o.newValue)
+	s.Noticef("Reloaded: MQTT max_ack_pending = %v", o.newValue)
+}
+
+// Compares options and disconnects clients that are no longer listed in pinned certs. Lock must not be held.
+func (s *Server) recheckPinnedCerts(curOpts *Options, newOpts *Options) {
+	s.mu.Lock()
+	disconnectClients := []*client{}
+	protoToPinned := map[int]PinnedCertSet{}
+	if !reflect.DeepEqual(newOpts.TLSPinnedCerts, curOpts.TLSPinnedCerts) {
+		protoToPinned[NATS] = curOpts.TLSPinnedCerts
+	}
+	if !reflect.DeepEqual(newOpts.MQTT.TLSPinnedCerts, curOpts.MQTT.TLSPinnedCerts) {
+		protoToPinned[MQTT] = curOpts.MQTT.TLSPinnedCerts
+	}
+	if !reflect.DeepEqual(newOpts.Websocket.TLSPinnedCerts, curOpts.Websocket.TLSPinnedCerts) {
+		protoToPinned[WS] = curOpts.Websocket.TLSPinnedCerts
+	}
+	for _, c := range s.clients {
+		if c.kind != CLIENT {
+			continue
+		}
+		if pinned, ok := protoToPinned[c.clientType()]; ok {
+			if !c.matchesPinnedCert(pinned) {
+				disconnectClients = append(disconnectClients, c)
+			}
+		}
+	}
+	checkClients := func(kind int, clients map[uint64]*client, set PinnedCertSet) {
+		for _, c := range clients {
+			if c.kind == kind && !c.matchesPinnedCert(set) {
+				disconnectClients = append(disconnectClients, c)
+			}
+		}
+	}
+	if !reflect.DeepEqual(newOpts.LeafNode.TLSPinnedCerts, curOpts.LeafNode.TLSPinnedCerts) {
+		checkClients(LEAF, s.leafs, newOpts.LeafNode.TLSPinnedCerts)
+	}
+	if !reflect.DeepEqual(newOpts.Cluster.TLSPinnedCerts, curOpts.Cluster.TLSPinnedCerts) {
+		checkClients(ROUTER, s.routes, newOpts.Cluster.TLSPinnedCerts)
+	}
+	if reflect.DeepEqual(newOpts.Gateway.TLSPinnedCerts, curOpts.Gateway.TLSPinnedCerts) {
+		for _, c := range s.remotes {
+			if !c.matchesPinnedCert(newOpts.Gateway.TLSPinnedCerts) {
+				disconnectClients = append(disconnectClients, c)
+			}
+		}
+	}
+	s.mu.Unlock()
+	if len(disconnectClients) > 0 {
+		s.Noticef("Disconnect %d clients due to pinned certs reload", len(disconnectClients))
+		for _, c := range disconnectClients {
+			c.closeConnection(TLSHandshakeError)
+		}
+	}
+}
+
+// Reload reads the current configuration file and calls out to ReloadOptions
+// to apply the changes. This returns an error if the server was not started
+// with a config file or an option which doesn't support hot-swapping was changed.
 func (s *Server) Reload() error {
 	s.mu.Lock()
-	if s.configFile == "" {
-		s.mu.Unlock()
+	configFile := s.configFile
+	s.mu.Unlock()
+	if configFile == "" {
 		return errors.New("can only reload config when a file is provided using -c or --config")
 	}
 
-	newOpts, err := ProcessConfigFile(s.configFile)
+	newOpts, err := ProcessConfigFile(configFile)
 	if err != nil {
-		s.mu.Unlock()
 		// TODO: Dump previous good config to a .bak file?
 		return err
 	}
+	return s.ReloadOptions(newOpts)
+}
+
+// ReloadOptions applies any supported options from the provided Option
+// type. This returns an error if an option which doesn't support
+// hot-swapping was changed.
+func (s *Server) ReloadOptions(newOpts *Options) error {
+	s.mu.Lock()
+
+	s.reloading = true
+	defer func() {
+		s.mu.Lock()
+		s.reloading = false
+		s.mu.Unlock()
+	}()
 
 	curOpts := s.getOpts()
 
@@ -597,6 +744,8 @@ func (s *Server) Reload() error {
 	clusterOrgPort := curOpts.Cluster.Port
 	gatewayOrgPort := curOpts.Gateway.Port
 	leafnodesOrgPort := curOpts.LeafNode.Port
+	websocketOrgPort := curOpts.Websocket.Port
+	mqttOrgPort := curOpts.MQTT.Port
 
 	s.mu.Unlock()
 
@@ -626,17 +775,25 @@ func (s *Server) Reload() error {
 	if newOpts.LeafNode.Port == -1 {
 		newOpts.LeafNode.Port = leafnodesOrgPort
 	}
+	if newOpts.Websocket.Port == -1 {
+		newOpts.Websocket.Port = websocketOrgPort
+	}
+	if newOpts.MQTT.Port == -1 {
+		newOpts.MQTT.Port = mqttOrgPort
+	}
 
 	if err := s.reloadOptions(curOpts, newOpts); err != nil {
 		return err
 	}
+
+	s.recheckPinnedCerts(curOpts, newOpts)
+
 	s.mu.Lock()
-	s.configTime = time.Now()
+	s.configTime = time.Now().UTC()
 	s.updateVarzConfigReloadableFields(s.varz)
 	s.mu.Unlock()
 	return nil
 }
-
 func applyBoolFlags(newOpts, flagOpts *Options) {
 	// Reset fields that may have been set to `true` in
 	// MergeOptions() when some of the flags default to `true`
@@ -715,9 +872,7 @@ func imposeOrder(value interface{}) error {
 			return value[i].String() < value[j].String()
 		})
 	case []string:
-		sort.Slice(value, func(i, j int) bool {
-			return value[i] < value[j]
-		})
+		sort.Strings(value)
 	case []*jwt.OperatorClaims:
 		sort.Slice(value, func(i, j int) bool {
 			return value[i].Issuer < value[j].Issuer
@@ -726,8 +881,11 @@ func imposeOrder(value interface{}) error {
 		sort.Slice(value.Gateways, func(i, j int) bool {
 			return value.Gateways[i].Name < value.Gateways[j].Name
 		})
-	case string, bool, int, int32, int64, time.Duration, float64, nil,
-		LeafNodeOpts, ClusterOpts, *tls.Config, *URLAccResolver, *MemAccResolver, Authentication:
+	case WebsocketOpts:
+		sort.Strings(value.AllowedOrigins)
+	case string, bool, uint8, int, int32, int64, time.Duration, float64, nil, LeafNodeOpts, ClusterOpts, *tls.Config, PinnedCertSet,
+		*URLAccResolver, *MemAccResolver, *DirAccResolver, *CacheDirAccResolver, Authentication, MQTTOpts, jwt.TagList,
+		*OCSPConfig, map[string]string, JSLimitOpts:
 		// explicitly skipped types
 	default:
 		// this will fail during unit tests
@@ -745,6 +903,14 @@ func (s *Server) diffOptions(newOpts *Options) ([]option, error) {
 		oldConfig = reflect.ValueOf(s.getOpts()).Elem()
 		newConfig = reflect.ValueOf(newOpts).Elem()
 		diffOpts  = []option{}
+
+		// Need to keep track of whether JS is being disabled
+		// to prevent changing limits at runtime.
+		jsEnabled           = s.JetStreamEnabled()
+		disableJS           bool
+		jsMemLimitsChanged  bool
+		jsFileLimitsChanged bool
+		jsStoreDirChanged   bool
 	)
 	for i := 0; i < oldConfig.NumField(); i++ {
 		field := oldConfig.Type().Field(i)
@@ -763,10 +929,17 @@ func (s *Server) diffOptions(newOpts *Options) ([]option, error) {
 		if err := imposeOrder(newValue); err != nil {
 			return nil, err
 		}
+
 		if changed := !reflect.DeepEqual(oldValue, newValue); !changed {
+			// Check to make sure we are running JetStream if we think we should be.
+			if strings.ToLower(field.Name) == "jetstream" && newValue.(bool) {
+				if !jsEnabled {
+					diffOpts = append(diffOpts, &jetStreamOption{newValue: true})
+				}
+			}
 			continue
 		}
-		switch strings.ToLower(field.Name) {
+		switch optName := strings.ToLower(field.Name); optName {
 		case "traceverbose":
 			diffOpts = append(diffOpts, &traceVerboseOption{newValue: newValue.(bool)})
 		case "trace":
@@ -785,6 +958,8 @@ func (s *Server) diffOptions(newOpts *Options) ([]option, error) {
 			diffOpts = append(diffOpts, &tlsOption{newValue: newValue.(*tls.Config)})
 		case "tlstimeout":
 			diffOpts = append(diffOpts, &tlsTimeoutOption{newValue: newValue.(float64)})
+		case "tlspinnedcerts":
+			diffOpts = append(diffOpts, &tlsPinnedCertOption{newValue: newValue.(PinnedCertSet)})
 		case "username":
 			diffOpts = append(diffOpts, &usernameOption{})
 		case "password":
@@ -854,6 +1029,15 @@ func (s *Server) diffOptions(newOpts *Options) ([]option, error) {
 			tmpNew := newValue.(GatewayOpts)
 			tmpOld.TLSConfig = nil
 			tmpNew.TLSConfig = nil
+			tmpOld.tlsConfigOpts = nil
+			tmpNew.tlsConfigOpts = nil
+
+			// Need to do the same for remote gateways' TLS configs.
+			// But we can't just set remotes' TLSConfig to nil otherwise this
+			// would lose the real TLS configuration.
+			tmpOld.Gateways = copyRemoteGWConfigsWithoutTLSConfig(tmpOld.Gateways)
+			tmpNew.Gateways = copyRemoteGWConfigsWithoutTLSConfig(tmpNew.Gateways)
+
 			// If there is really a change prevents reload.
 			if !reflect.DeepEqual(tmpOld, tmpNew) {
 				// See TODO(ik) note below about printing old/new values.
@@ -866,12 +1050,147 @@ func (s *Server) diffOptions(newOpts *Options) ([]option, error) {
 			tmpNew := newValue.(LeafNodeOpts)
 			tmpOld.TLSConfig = nil
 			tmpNew.TLSConfig = nil
+			tmpOld.tlsConfigOpts = nil
+			tmpNew.tlsConfigOpts = nil
+
+			// Need to do the same for remote leafnodes' TLS configs.
+			// But we can't just set remotes' TLSConfig to nil otherwise this
+			// would lose the real TLS configuration.
+			tmpOld.Remotes = copyRemoteLNConfigForReloadCompare(tmpOld.Remotes)
+			tmpNew.Remotes = copyRemoteLNConfigForReloadCompare(tmpNew.Remotes)
+
+			// Special check for leafnode remotes changes which are not supported right now.
+			leafRemotesChanged := func(a, b LeafNodeOpts) bool {
+				if len(tmpOld.Remotes) != len(tmpNew.Remotes) {
+					return true
+				}
+
+				// Check whether all remotes URLs are still the same.
+				for _, oldRemote := range tmpOld.Remotes {
+					var found bool
+
+					if oldRemote.LocalAccount == _EMPTY_ {
+						oldRemote.LocalAccount = globalAccountName
+					}
+
+					for _, newRemote := range tmpNew.Remotes {
+						// Bind to global account in case not defined.
+						if newRemote.LocalAccount == _EMPTY_ {
+							newRemote.LocalAccount = globalAccountName
+						}
+
+						if reflect.DeepEqual(oldRemote, newRemote) {
+							found = true
+							break
+						}
+					}
+
+					if !found {
+						return true
+					}
+				}
+
+				return false
+			}
+
+			// First check whether remotes changed at all. If they did not,
+			// skip them in the complete equal check.
+			if !leafRemotesChanged(tmpOld, tmpNew) {
+				tmpOld.Remotes = nil
+				tmpNew.Remotes = nil
+			}
+
 			// If there is really a change prevents reload.
 			if !reflect.DeepEqual(tmpOld, tmpNew) {
 				// See TODO(ik) note below about printing old/new values.
 				return nil, fmt.Errorf("config reload not supported for %s: old=%v, new=%v",
 					field.Name, oldValue, newValue)
 			}
+		case "jetstream":
+			new := newValue.(bool)
+			old := oldValue.(bool)
+			if new != old {
+				diffOpts = append(diffOpts, &jetStreamOption{newValue: new})
+			}
+
+			// Mark whether JS will be disabled.
+			disableJS = !new
+		case "storedir":
+			new := newValue.(string)
+			old := oldValue.(string)
+			modified := new != old
+
+			// Check whether JS is being disabled and/or storage dir attempted to change.
+			if jsEnabled && modified {
+				if new == _EMPTY_ {
+					// This means that either JS is being disabled or it is using an temp dir.
+					// Allow the change but error in case JS was not disabled.
+					jsStoreDirChanged = true
+				} else {
+					return nil, fmt.Errorf("config reload not supported for jetstream storage directory")
+				}
+			}
+		case "jetstreammaxmemory", "jetstreammaxstore":
+			old := oldValue.(int64)
+			new := newValue.(int64)
+
+			// Check whether JS is being disabled and/or limits are being changed.
+			var (
+				modified  = new != old
+				fromUnset = old == -1
+				fromSet   = !fromUnset
+				toUnset   = new == -1
+				toSet     = !toUnset
+			)
+			if jsEnabled && modified {
+				// Cannot change limits from dynamic storage at runtime.
+				switch {
+				case fromSet && toUnset:
+					// Limits changed but it may mean that JS is being disabled,
+					// keep track of the change and error in case it is not.
+					switch optName {
+					case "jetstreammaxmemory":
+						jsMemLimitsChanged = true
+					case "jetstreammaxstore":
+						jsFileLimitsChanged = true
+					default:
+						return nil, fmt.Errorf("config reload not supported for jetstream max memory and store")
+					}
+				case fromUnset && toSet:
+					// Prevent changing from dynamic max memory / file at runtime.
+					return nil, fmt.Errorf("config reload not supported for jetstream dynamic max memory and store")
+				default:
+					return nil, fmt.Errorf("config reload not supported for jetstream max memory and store")
+				}
+			}
+		case "websocket":
+			// Similar to gateways
+			tmpOld := oldValue.(WebsocketOpts)
+			tmpNew := newValue.(WebsocketOpts)
+			tmpOld.TLSConfig = nil
+			tmpNew.TLSConfig = nil
+			// If there is really a change prevents reload.
+			if !reflect.DeepEqual(tmpOld, tmpNew) {
+				// See TODO(ik) note below about printing old/new values.
+				return nil, fmt.Errorf("config reload not supported for %s: old=%v, new=%v",
+					field.Name, oldValue, newValue)
+			}
+		case "mqtt":
+			diffOpts = append(diffOpts, &mqttAckWaitReload{newValue: newValue.(MQTTOpts).AckWait})
+			diffOpts = append(diffOpts, &mqttMaxAckPendingReload{newValue: newValue.(MQTTOpts).MaxAckPending})
+			// Nil out/set to 0 the options that we allow to be reloaded so that
+			// we only fail reload if some that we don't support are changed.
+			tmpOld := oldValue.(MQTTOpts)
+			tmpNew := newValue.(MQTTOpts)
+			tmpOld.TLSConfig, tmpOld.AckWait, tmpOld.MaxAckPending = nil, 0, 0
+			tmpNew.TLSConfig, tmpNew.AckWait, tmpNew.MaxAckPending = nil, 0, 0
+			if !reflect.DeepEqual(tmpOld, tmpNew) {
+				// See TODO(ik) note below about printing old/new values.
+				return nil, fmt.Errorf("config reload not supported for %s: old=%v, new=%v",
+					field.Name, oldValue, newValue)
+			}
+			tmpNew.AckWait = newValue.(MQTTOpts).AckWait
+			tmpNew.MaxAckPending = newValue.(MQTTOpts).MaxAckPending
 		case "connecterrorreports":
 			diffOpts = append(diffOpts, &connectErrorReports{newValue: newValue.(int)})
 		case "reconnecterrorreports":
@@ -892,6 +1211,25 @@ func (s *Server) diffOptions(newOpts *Options) ([]option, error) {
 				continue
 			}
 			fallthrough
+		case "noauthuser":
+			if oldValue != _EMPTY_ && newValue == _EMPTY_ {
+				for _, user := range newOpts.Users {
+					if user.Username == oldValue {
+						return nil, fmt.Errorf("config reload not supported for %s: old=%v, new=%v",
+							field.Name, oldValue, newValue)
+					}
+				}
+			} else {
+				return nil, fmt.Errorf("config reload not supported for %s: old=%v, new=%v",
+					field.Name, oldValue, newValue)
+			}
+		case "systemaccount":
+			if oldValue != DEFAULT_SYSTEM_ACCOUNT || newValue != _EMPTY_ {
+				return nil, fmt.Errorf("config reload not supported for %s: old=%v, new=%v",
+					field.Name, oldValue, newValue)
+			}
+		case "ocspconfig":
+			diffOpts = append(diffOpts, &ocspOption{newValue: newValue.(*OCSPConfig)})
 		default:
 			// TODO(ik): Implement String() on those options to have a nice print.
 			// %v is difficult to figure what's what, %+v print private fields and
@@ -904,15 +1242,64 @@ func (s *Server) diffOptions(newOpts *Options) ([]option, error) {
 		}
 	}
 
+	// If not disabling JS but limits have changed then it is an error.
+	if !disableJS {
+		if jsMemLimitsChanged || jsFileLimitsChanged {
+			return nil, fmt.Errorf("config reload not supported for jetstream max memory and max store")
+		}
+		if jsStoreDirChanged {
+			return nil, fmt.Errorf("config reload not supported for jetstream storage dir")
+		}
+	}
+
 	return diffOpts, nil
 }
 
+func copyRemoteGWConfigsWithoutTLSConfig(current []*RemoteGatewayOpts) []*RemoteGatewayOpts {
+	l := len(current)
+	if l == 0 {
+		return nil
+	}
+	rgws := make([]*RemoteGatewayOpts, 0, l)
+	for _, rcfg := range current {
+		cp := *rcfg
+		cp.TLSConfig = nil
+		cp.tlsConfigOpts = nil
+		rgws = append(rgws, &cp)
+	}
+	return rgws
+}
+
+func copyRemoteLNConfigForReloadCompare(current []*RemoteLeafOpts) []*RemoteLeafOpts {
+	l := len(current)
+	if l == 0 {
+		return nil
+	}
+	rlns := make([]*RemoteLeafOpts, 0, l)
+	for _, rcfg := range current {
+		cp := *rcfg
+		cp.TLSConfig = nil
+		cp.tlsConfigOpts = nil
+		// This is set only when processing a CONNECT, so reset here so that we
+		// don't fail the DeepEqual comparison.
+		cp.TLS = false
+		// For now, remove DenyImports/Exports since those get modified at runtime
+		// to add JS APIs.
+		cp.DenyImports, cp.DenyExports = nil, nil
+		rlns = append(rlns, &cp)
+	}
+	return rlns
+}
+
 func (s *Server) applyOptions(ctx *reloadContext, opts []option) {
 	var (
 		reloadLogging      = false
 		reloadAuth         = false
 		reloadClusterPerms = false
 		reloadClientTrcLvl = false
+		reloadJetstream    = false
+		jsEnabled          = false
+		reloadTLS          = false
 	)
 	for _, opt := range opts {
 		opt.Apply(s)
@@ -925,9 +1312,16 @@ func (s *Server) applyOptions(ctx *reloadContext, opts []option) {
 		if opt.IsAuthChange() {
 			reloadAuth = true
 		}
+		if opt.IsTLSChange() {
+			reloadTLS = true
+		}
 		if opt.IsClusterPermsChange() {
 			reloadClusterPerms = true
 		}
+		if opt.IsJetStreamChange() {
+			reloadJetstream = true
+			jsEnabled = opt.(*jetStreamOption).newValue
+		}
 	}
 
 	if reloadLogging {
@@ -943,9 +1337,55 @@ func (s *Server) applyOptions(ctx *reloadContext, opts []option) {
 		s.reloadClusterPermissions(ctx.oldClusterPerms)
 	}
 
+	if reloadJetstream {
+		if !jsEnabled {
+			s.DisableJetStream()
+		} else if !s.JetStreamEnabled() {
+			if err := s.restartJetStream(); err != nil {
+				s.Warnf("Can't start JetStream: %v", err)
+			}
+		}
+		// Make sure to reset the internal loop's version of JS.
+		s.resetInternalLoopInfo()
+		s.sendStatszUpdate()
+	}
+
+	// For remote gateways and leafnodes, make sure that their TLS configuration
+	// is updated (since the config is "captured" early and changes would otherwise
+	// not be visible).
+	newOpts := s.getOpts()
+	if s.gateway.enabled {
+		s.gateway.updateRemotesTLSConfig(newOpts)
+	}
+	if len(newOpts.LeafNode.Remotes) > 0 {
+		s.updateRemoteLeafNodesTLSConfig(newOpts)
+	}
+
+	if reloadTLS {
+		// Restart OCSP monitoring.
+		if err := s.reloadOCSP(); err != nil {
+			s.Warnf("Can't restart OCSP Stapling: %v", err)
+		}
+	}
+
 	s.Noticef("Reloaded server configuration")
 }
 
+// This will send a reset to the internal send loop.
+func (s *Server) resetInternalLoopInfo() {
+	var resetCh chan struct{}
+	s.mu.Lock()
+	if s.sys != nil {
+		// can't hold the lock as go routine reading it may be waiting for lock as well
+		resetCh = s.sys.resetCh
+	}
+	s.mu.Unlock()
+
+	if resetCh != nil {
+		resetCh <- struct{}{}
+	}
+}
+
 // Update all cached debug and trace settings for every client
 func (s *Server) reloadClientTraceLevel() {
 	opts := s.getOpts()
@@ -1002,10 +1442,11 @@ func (s *Server) reloadAuthorization() {
 	// This map will contain the names of accounts that have their streams
 	// import configuration changed.
 	awcsti := make(map[string]struct{})
-
+	checkJetStream := false
 	s.mu.Lock()
 
-	// This can not be changed for now so ok to check server's trustedKeys
+	// This can not be changed for now so ok to check server's trustedKeys unlocked.
+	// If plain configured accounts, process here.
 	if s.trustedKeys == nil {
 		// We need to drain the old accounts here since we have something
 		// new configured. We do not want s.accounts to change since that would
@@ -1014,15 +1455,19 @@ func (s *Server) reloadAuthorization() {
 		oldAccounts := make(map[string]*Account)
 		s.accounts.Range(func(k, v interface{}) bool {
 			acc := v.(*Account)
-			acc.mu.RLock()
+			acc.mu.Lock()
 			oldAccounts[acc.Name] = acc
-			acc.mu.RUnlock()
+			// Need to clear out eventing timers since they close over this account and not the new one.
+			clearTimer(&acc.etmr)
+			clearTimer(&acc.ctmr)
+			acc.mu.Unlock()
 			s.accounts.Delete(k)
 			return true
 		})
 		s.gacc = nil
 		s.configureAccounts()
 		s.configureAuthorization()
+		s.mu.Unlock()
 
 		s.accounts.Range(func(k, v interface{}) bool {
 			newAcc := v.(*Account)
@@ -1030,15 +1475,35 @@ func (s *Server) reloadAuthorization() {
 				// If account exist in latest config, "transfer" the account's
 				// sublist and client map to the new account.
 				acc.mu.RLock()
+				newAcc.mu.Lock()
 				if len(acc.clients) > 0 {
 					newAcc.clients = make(map[*client]struct{}, len(acc.clients))
 					for c := range acc.clients {
 						newAcc.clients[c] = struct{}{}
 					}
 				}
+
 				newAcc.sl = acc.sl
 				newAcc.rm = acc.rm
-				newAcc.respMap = acc.respMap
+				// Transfer internal client state. The configureAccounts call from above may have set up a new one.
+				// We need to use the old one, and the isid to not confuse internal subs.
+				newAcc.ic, newAcc.isid = acc.ic, acc.isid
+				// Transfer any JetStream state.
+				newAcc.js = acc.js
+				// Also transfer any internal accounting on different client types. We copy over all clients
+				// so need to copy this as well for proper accounting going forward.
+				newAcc.nrclients = acc.nrclients
+				newAcc.sysclients = acc.sysclients
+				newAcc.nleafs = acc.nleafs
+				newAcc.nrleafs = acc.nrleafs
+				// Process any reverse map entries.
+				if len(acc.imports.rrMap) > 0 {
+					newAcc.imports.rrMap = make(map[string][]*serviceRespEntry)
+					for k, v := range acc.imports.rrMap {
+						newAcc.imports.rrMap[k] = v
+					}
+				}
+				newAcc.mu.Unlock()
 				acc.mu.RUnlock()
 
 				// Check if current and new config of this account are same
@@ -1046,9 +1511,20 @@ func (s *Server) reloadAuthorization() {
 				if !acc.checkStreamImportsEqual(newAcc) {
 					awcsti[newAcc.Name] = struct{}{}
 				}
+
+				// We need to remove all old service import subs.
+				acc.removeAllServiceImportSubs()
+				newAcc.addAllServiceImportSubs()
 			}
 			return true
 		})
+		s.mu.Lock()
+		// Check if we had a default system account.
+		if s.sys != nil && s.sys.account != nil && !s.opts.NoSystemAccount {
+			s.accounts.Store(s.sys.account.Name, s.sys.account)
+		}
+		// Double check any JetStream configs.
+		checkJetStream = s.js != nil
 	} else if s.opts.AccountResolver != nil {
 		s.configureResolver()
 		if _, ok := s.accResolver.(*MemAccResolver); ok {
@@ -1063,15 +1539,12 @@ func (s *Server) reloadAuthorization() {
 				if acc == s.gacc {
 					return true
 				}
-				acc.mu.RLock()
-				accName := acc.Name
-				acc.mu.RUnlock()
+				accName := acc.GetName()
 				// Release server lock for following actions
 				s.mu.Unlock()
 				accClaims, claimJWT, _ := s.fetchAccountClaims(accName)
 				if accClaims != nil {
-					err := s.updateAccountWithClaimJWT(acc, claimJWT)
-					if err != nil && err != ErrAccountResolverSameClaims {
+					if err := s.updateAccountWithClaimJWT(acc, claimJWT); err != nil {
 						s.Noticef("Reloaded: deleting account [bad claims]: %q", accName)
 						s.accounts.Delete(k)
 					}
@@ -1086,8 +1559,6 @@ func (s *Server) reloadAuthorization() {
 		}
 	}
 
-	// Gather clients that changed accounts. We will close them and they
-	// will reconnect, doing the right thing.
 	var (
 		cclientsa [64]*client
 		cclients  = cclientsa[:0]
@@ -1096,6 +1567,9 @@ func (s *Server) reloadAuthorization() {
 		routesa   [64]*client
 		routes    = routesa[:0]
 	)
+
+	// Gather clients that changed accounts. We will close them and they
+	// will reconnect, doing the right thing.
 	for _, client := range s.clients {
 		if s.clientHasMovedToDifferentAccount(client) {
 			cclients = append(cclients, client)
@@ -1106,6 +1580,26 @@ func (s *Server) reloadAuthorization() {
 	for _, route := range s.routes {
 		routes = append(routes, route)
 	}
+	// Check here for any system/internal clients which will not be in the servers map of normal clients.
+	if s.sys != nil && s.sys.account != nil && !s.opts.NoSystemAccount {
+		s.accounts.Store(s.sys.account.Name, s.sys.account)
+	}
+
+	s.accounts.Range(func(k, v interface{}) bool {
+		acc := v.(*Account)
+		acc.mu.RLock()
+		// Check for sysclients accounting, ignore the system account.
+		if acc.sysclients > 0 && (s.sys == nil || s.sys.account != acc) {
+			for c := range acc.clients {
+				if c.kind != CLIENT && c.kind != LEAF {
+					clients = append(clients, c)
+				}
+			}
+		}
+		acc.mu.RUnlock()
+		return true
+	})
+
 	var resetCh chan struct{}
 	if s.sys != nil {
 		// can't hold the lock as go routine reading it may be waiting for lock as well
@@ -1117,19 +1611,25 @@ func (s *Server) reloadAuthorization() {
 		resetCh <- struct{}{}
 	}
 
+	// Check that publish retained messages sources are still allowed to publish.
+	s.mqttCheckPubRetainedPerms()
+
 	// Close clients that have moved accounts
 	for _, client := range cclients {
 		client.closeConnection(ClientClosed)
 	}
 
-	for _, client := range clients {
+	for _, c := range clients {
 		// Disconnect any unauthorized clients.
-		if !s.isClientAuthorized(client) {
-			client.authViolation()
+		// Ignore internal clients.
+		if (c.kind == CLIENT || c.kind == LEAF) && !s.isClientAuthorized(c) {
+			c.authViolation()
 			continue
 		}
+		// Check to make sure account is correct.
+		c.swapAccountAfterReload()
 		// Remove any unauthorized subscriptions and check for account imports.
-		client.processSubsOnConfigReload(awcsti)
+		c.processSubsOnConfigReload(awcsti)
 	}
 
 	for _, route := range routes {
@@ -1142,11 +1642,22 @@ func (s *Server) reloadAuthorization() {
 			route.authViolation()
 		}
 	}
+
+	if res := s.AccountResolver(); res != nil {
+		res.Reload()
+	}
+
+	// We will double check all JetStream configs on a reload.
+	if checkJetStream {
+		if err := s.enableJetStreamAccounts(); err != nil {
+			s.Errorf(err.Error())
+		}
+	}
 }
 
 // Returns true if given client current account has changed (or user
 // no longer exist) in the new config, false if the user did not
-// change account.
+// change accounts.
 // Server lock is held on entry.
 func (s *Server) clientHasMovedToDifferentAccount(c *client) bool {
 	var (
@@ -1245,7 +1756,7 @@ func (s *Server) reloadClusterPermissions(oldPerms *RoutePermissions) {
 		deleteRoutedSubs []*subscription
 	)
 	// FIXME(dlc) - Change for accounts.
-	gacc.sl.localSubs(&localSubs)
+	gacc.sl.localSubs(&localSubs, false)
 
 	// Go through all local subscriptions
 	for _, sub := range localSubs {
