@@ -1,4 +1,4 @@
-// Copyright 2019 The NATS Authors
+// Copyright 2019-2020 The NATS Authors
 // Licensed under the Apache License, Version 2.0 (the "License");
 // you may not use this file except in compliance with the License.
 // You may obtain a copy of the License at
@@ -11,6 +11,7 @@
 // See the License for the specific language governing permissions and
 // limitations under the License.
 
+//go:build !race
 // +build !race
 
 package test
@@ -20,9 +21,9 @@ import (
 	"encoding/json"
 	"fmt"
 	"io/ioutil"
+	"math/rand"
 	"net"
 	"net/url"
-	"os"
 	"regexp"
 	"runtime"
 	"strconv"
@@ -47,9 +48,10 @@ func TestNoRaceRouteSendSubs(t *testing.T) {
 				port: -1
 				%s
 			}
+			no_sys_acc: true
 	`
 	cfa := createConfFile(t, []byte(fmt.Sprintf(template, "")))
-	defer os.Remove(cfa)
+	defer removeFile(t, cfa)
 	srvA, optsA := RunServerWithConfig(cfa)
 	srvA.Shutdown()
 	optsA.DisableShortFirstPing = true
@@ -79,8 +81,11 @@ func TestNoRaceRouteSendSubs(t *testing.T) {
 
 	// total number of subscriptions per server
 	totalPerServer := 100000
-	for i := 0; i < totalPerServer; i++ {
-		proto := fmt.Sprintf("SUB foo.%d %d\r\n", i, i*2)
+	for i := 0; i < totalPerServer/2; i++ {
+		proto := fmt.Sprintf("SUB foo.%d %d\r\n", i, i*2+1)
+		clientASend(proto)
+		clientBSend(proto)
+		proto = fmt.Sprintf("SUB bar.%d queue.%d %d\r\n", i, i, i*2+2)
 		clientASend(proto)
 		clientBSend(proto)
 	}
@@ -326,11 +331,11 @@ func TestNoRaceLargeClusterMem(t *testing.T) {
 	checkClusterFormed(t, servers...)
 
 	// Calculate in MB what we are using now.
-	const max = 50 * 1024 * 1024 // 50MB
+	const max = 80 * 1024 * 1024 // 80MB
 	runtime.ReadMemStats(&m)
 	used := m.TotalAlloc - pta
 	if used > max {
-		t.Fatalf("Cluster using too much memory, expect < 50MB, got %dMB", used/(1024*1024))
+		t.Fatalf("Cluster using too much memory, expect < 80MB, got %dMB", used/(1024*1024))
 	}
 
 	for _, s := range servers {
@@ -370,11 +375,18 @@ func TestNoRaceQueueSubWeightOrderMultipleConnections(t *testing.T) {
 	info := checkInfoMsg(t, rc)
 
 	info.ID = routeID
+	info.Name = routeID
+
 	b, err := json.Marshal(info)
 	if err != nil {
 		t.Fatalf("Could not marshal test route info: %v", err)
 	}
-	routeSend(fmt.Sprintf("INFO %s\r\n", b))
+	// Send our INFO and wait for a PONG. This will prevent a race
+	// where server will have started processing queue subscriptions
+	// and then processes the route's INFO (sending its current subs)
+	// followed by updates.
+	routeSend(fmt.Sprintf("INFO %s\r\nPING\r\n", b))
+	routeExpect(pongRe)
 
 	start := make(chan bool)
 	for _, nc := range clients {
@@ -382,7 +394,7 @@ func TestNoRaceQueueSubWeightOrderMultipleConnections(t *testing.T) {
 			<-start
 			// Now create 100 identical queue subscribers on each connection.
 			for i := 0; i < 100; i++ {
-				if _, err := nc.QueueSubscribeSync("foo", "bar"); err != nil {
+				if _, err := nc.QueueSubscribe("foo", "bar", func(_ *nats.Msg) {}); err != nil {
 					return
 				}
 			}
@@ -434,10 +446,14 @@ func TestNoRaceClusterLeaksSubscriptions(t *testing.T) {
 	numResponses := 100
 	repliers := make([]*nats.Conn, 0, numResponses)
 
+	var noOpErrHandler = func(_ *nats.Conn, _ *nats.Subscription, _ error) {}
+
 	// Create 100 repliers
 	for i := 0; i < 50; i++ {
 		nc1, _ := nats.Connect(urlA)
+		nc1.SetErrorHandler(noOpErrHandler)
 		nc2, _ := nats.Connect(urlB)
+		nc2.SetErrorHandler(noOpErrHandler)
 		repliers = append(repliers, nc1, nc2)
 		nc1.Subscribe("test.reply", func(m *nats.Msg) {
 			m.Respond([]byte("{\"sender\": 22 }"))
@@ -461,11 +477,13 @@ func TestNoRaceClusterLeaksSubscriptions(t *testing.T) {
 		// Create 8 queue Subscribers for responses.
 		for i := 0; i < 8; i++ {
 			nc, _ := nats.Connect(servers)
+			nc.SetErrorHandler(noOpErrHandler)
 			nc.ChanQueueSubscribe(inbox, grp, msgs)
 			nc.Flush()
 			defer nc.Close()
 		}
 		nc, _ := nats.Connect(servers)
+		nc.SetErrorHandler(noOpErrHandler)
 		nc.PublishRequest("test.reply", inbox, req)
 		defer nc.Close()
 
@@ -626,3 +644,61 @@ func TestNoRaceLeafNodeSmapUpdate(t *testing.T) {
 	// Expect that many LS-
 	checkLS("LS- ", ns)
 }
+
+func TestNoRaceSlowProxy(t *testing.T) {
+	t.Skip()
+
+	opts := DefaultTestOptions
+	opts.Port = -1
+	s := RunServer(&opts)
+	defer s.Shutdown()
+
+	rttTarget := 22 * time.Millisecond
+	bwTarget := 10 * 1024 * 1024 / 8 // 10mbit
+
+	sp := newSlowProxy(rttTarget, bwTarget, bwTarget, &opts)
+	defer sp.stop()
+
+	nc, err := nats.Connect(sp.clientURL())
+	if err != nil {
+		t.Fatalf("Unexpected error: %v", err)
+	}
+	defer nc.Close()
+
+	doRTT := func() time.Duration {
+		t.Helper()
+		const samples = 5
+		var total time.Duration
+		for i := 0; i < samples; i++ {
+			rtt, _ := nc.RTT()
+			total += rtt
+		}
+		return total / samples
+	}
+
+	rtt := doRTT()
+	if rtt < rttTarget || rtt > (rttTarget*3/2) {
+		t.Fatalf("rtt is out of range, target of %v, actual %v", rttTarget, rtt)
+	}
+
+	// Now test send BW.
+	const payloadSize = 64 * 1024
+	var payload [payloadSize]byte
+	rand.Read(payload[:])
+
+	// 5MB total.
+	bytesSent := (5 * 1024 * 1024)
+	toSend := bytesSent / payloadSize
+
+	start := time.Now()
+	for i := 0; i < toSend; i++ {
+		nc.Publish("z", payload[:])
+	}
+	nc.Flush()
+	tt := time.Since(start)
+	bps := float64(bytesSent) / tt.Seconds()
+	min, max := float64(bwTarget)*0.8, float64(bwTarget)*1.25
+	if bps < min || bps > max {
+		t.Fatalf("bps is off, target is %v, actual is %v", bwTarget, bps)
+	}
+}
