@@ -18,13 +18,13 @@ import (
 	"fmt"
 	"io/ioutil"
 	"net"
-	"os"
 	"runtime"
 	"strconv"
 	"sync"
 	"testing"
 	"time"
 
+	"github.com/nats-io/nats-server/v2/internal/testhelper"
 	"github.com/nats-io/nats-server/v2/server"
 	"github.com/nats-io/nats.go"
 )
@@ -35,6 +35,10 @@ func runRouteServer(t *testing.T) (*server.Server, *server.Options) {
 	return RunServerWithConfig("./configs/cluster.conf")
 }
 
+func runRouteServerOverrides(t *testing.T, cbo func(*server.Options), cbs func(*server.Server)) (*server.Server, *server.Options) {
+	return RunServerWithConfigOverrides("./configs/cluster.conf", cbo, cbs)
+}
+
 func TestRouterListeningSocket(t *testing.T) {
 	s, opts := runRouteServer(t)
 	defer s.Shutdown()
@@ -82,6 +86,7 @@ func TestSendRouteInfoOnConnect(t *testing.T) {
 	// Need to send a different INFO than the one received, otherwise the server
 	// will detect as a "cycle" and close the connection.
 	info.ID = routeID
+	info.Name = ""
 	b, err := json.Marshal(info)
 	if err != nil {
 		t.Fatalf("Could not marshal test route info: %v", err)
@@ -93,7 +98,11 @@ func TestSendRouteInfoOnConnect(t *testing.T) {
 }
 
 func TestRouteToSelf(t *testing.T) {
-	s, opts := runRouteServer(t)
+	l := testhelper.NewDummyLogger(100)
+	s, opts := runRouteServerOverrides(t, nil,
+		func(s *server.Server) {
+			s.SetLogger(l, true, true)
+		})
 	defer s.Shutdown()
 
 	rc := createRouteConn(t, opts.Cluster.Host, opts.Cluster.Port)
@@ -123,6 +132,8 @@ func TestRouteToSelf(t *testing.T) {
 	if _, err := rc.Read(buf); err == nil {
 		t.Fatal("Expected route connection to be closed")
 	}
+	// This should have been removed by removePassFromTrace(), but we also check debug logs here
+	l.CheckForProhibited(t, "route authorization password found", "top_secret")
 }
 
 func TestSendRouteSubAndUnsub(t *testing.T) {
@@ -375,7 +386,11 @@ func TestRouteQueueSemantics(t *testing.T) {
 }
 
 func TestSolicitRouteReconnect(t *testing.T) {
-	s, opts := runRouteServer(t)
+	l := testhelper.NewDummyLogger(100)
+	s, opts := runRouteServerOverrides(t, nil,
+		func(s *server.Server) {
+			s.SetLogger(l, true, true)
+		})
 	defer s.Shutdown()
 
 	rURL := opts.Routes[0]
@@ -388,6 +403,9 @@ func TestSolicitRouteReconnect(t *testing.T) {
 	// We expect to get called back..
 	route = acceptRouteConn(t, rURL.Host, 2*server.DEFAULT_ROUTE_CONNECT)
 	route.Close()
+
+	// Now we want to check for the debug logs when it tries to reconnect
+	l.CheckForProhibited(t, "route authorization password found", ":bar")
 }
 
 func TestMultipleRoutesSameId(t *testing.T) {
@@ -478,6 +496,7 @@ func TestRouteResendsLocalSubsOnReconnect(t *testing.T) {
 		t.Fatalf("Could not unmarshal route info: %v", err)
 	}
 	info.ID = "ROUTE:1234"
+	info.Name = ""
 	b, err := json.Marshal(info)
 	if err != nil {
 		t.Fatalf("Could not marshal test route info: %v", err)
@@ -603,6 +622,7 @@ func TestRouteSendAsyncINFOToClients(t *testing.T) {
 		sendRouteINFO := func(routeSend sendFun, routeExpect expectFun, urls []string) {
 			routeInfo := server.Info{}
 			routeInfo.ID = routeID
+			routeInfo.Cluster = "xyz"
 			routeInfo.Host = "127.0.0.1"
 			routeInfo.Port = 5222
 			routeInfo.ClientConnectURLs = urls
@@ -1080,14 +1100,11 @@ func TestRouteBasicPermissions(t *testing.T) {
 
 func createConfFile(t *testing.T, content []byte) string {
 	t.Helper()
-	conf, err := ioutil.TempFile("", "")
-	if err != nil {
-		t.Fatalf("Error creating conf file: %v", err)
-	}
+	conf := createFile(t, "")
 	fName := conf.Name()
 	conf.Close()
 	if err := ioutil.WriteFile(fName, content, 0666); err != nil {
-		os.Remove(fName)
+		removeFile(t, fName)
 		t.Fatalf("Error writing conf file: %v", err)
 	}
 	return fName
@@ -1122,6 +1139,7 @@ func TestRoutesOnlyImportOrExport(t *testing.T) {
 		cf := createConfFile(t, []byte(fmt.Sprintf(`
 			port: -1
 			cluster {
+				name: "Z"
 				port: -1
 				authorization {
 					user: ivan
@@ -1132,7 +1150,7 @@ func TestRoutesOnlyImportOrExport(t *testing.T) {
 				}
 			}
 		`, c)))
-		defer os.Remove(cf)
+		defer removeFile(t, cf)
 		s, _ := RunServerWithConfig(cf)
 		s.Shutdown()
 	}
